#!/usr/bin/perl -w 
###############################################################################
#   afick_common.pl
#      it's a part of the afick project
#
#    Copyright (C) 2002-2004 by Eric Gerbier
#    Bug reports to: gerbier@users.sourceforge.net
#    $Id: afick-common.pl,v 1.5 2005/01/13 08:41:29 gerbier Exp $
#
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
###############################################################################
# common sub and var for afick's tool
###############################################################################
# I use special naming for references :
# $r_name : for a ref to scalar
# $rh_name : for a ref to hashes
# $ra_name : for a ref to array

use strict;

{

	# alias table
	# contains pre-defined values
	my %alias = (
		a => 'a',    # atime
		b => 'b',    # blocks
		c => 'c',    # ctime
		d => 'd',    # device
		g => 'g',    # gid
		i => 'i',    # inode
		m => 'm',    # mtime

		#	h    => '5',              # md5 hash by default (compatibility)
		md5  => '5',              # md5 hash
		sha1 => '1',              # sha1
		n    => 'n',              # number of links
		p    => 'p',              # permissions
		u    => 'u',              # uid
		s    => 's',              # size
		all  => 'pdinugsbmc5',    # all except a
		R    => 'pdinugsmc5',     # aide compatibility
		L    => 'pdinug',         # aide compatibility
		E    => ' ',              # aide compatibility
	);

	sub get_default_alias() {
		return %alias;
	}

}

###############################################################################
{

# a block to store error messages
# to be used like errno
# if a subroutine return undef, it set a message error to be retrieved by get_error
# (a call to get_error reset error message)

	# a "private" var (in a block)
	my $error = '';

	sub set_error($) {
		$error = shift(@_);
	}

	sub get_error() {
		my $msg = $error;

		# reset error
		$error = '';
		return $msg;
	}
}
###############################################################################
# return default config file according operating system
sub get_default_config() {
	my $default_config_file;

	if ( $^O =~ m/linux/ ) {
		$default_config_file = '/etc/afick.conf';
	}
	elsif ( $^O =~ m/^MSWin/ ) {
		$default_config_file = 'windows.conf';
	}
	else {
		$default_config_file = 'afick.conf';
	}

	return $default_config_file;
}
#######################################################
# sub debug
# sub warning
# are to be defined in each calling program
#######################################################
# remove_trailing blanks
# syntaxe : remove_trailing_spaces(\$var)
sub remove_trailing_spaces ($) {
	my $r_val = shift(@_);    # ref to var to clean

	$$r_val =~ s/^\s+//;
	$$r_val =~ s/\s+$//;

	return $$r_val;
}
#######################################################
# remove all spaces in a string
# syntaxe : remove_all_spaces(\$var)
sub remove_all_spaces ($) {
	my $r_val = shift(@_);

	$$r_val =~ s/\s+//g;

	return $$r_val;
}
###############################################################################
# read config file and have some cleaning
sub read_config($$) {
	my $configfile = shift(@_);    # config file name
	my $ra_config  = shift(@_);    # array to store config lines

	if ( open( CONFIG, '<', $configfile ) ) {
		while ( my $ligne = <CONFIG> ) {

			# normalize config line
			chomp($ligne);
			remove_trailing_spaces( \$ligne );
			push( @$ra_config, $ligne );
		}
		close(CONFIG);
		return 1;
	}
	else {
		set_error("WARNING: can not open $configfile : $!\n");
		return undef;
	}
}
###############################################################################
# write to a config file
sub write_config($$) {
	my $configfile = shift(@_);    # config file name
	my $ra_config  = shift(@_);    # array containing config lines

	if ( open( CONFIG, '>', $configfile ) ) {
		foreach my $ligne (@$ra_config) {
			print CONFIG $ligne . "\n";
		}
		close(CONFIG);
		return 1;
	}
	else {
		set_error("WARNING: can not write to $configfile : $!\n");
		return undef;
	}
}
###############################################################################
# patterns for following config line tests

my $macro_pattern     = '^\@\@define\s+(\w+)\s*(.*)';    # @@define name val
my $alias_pattern     = '^(\w+)\s*=\s*(.*)';             # name = val
my $directive_pattern = '^(\w+)\s*:=\s*(.*)';            # name := val

my $negsel_pattern   = '^!\s*(.*)';         # ! ligne
my $equalsel_pattern = '^=\s*(.*)';         # = ligne
my $anysel_pattern   = '^[=!]?\s*(.*)';     # [!/=] ligne
my $file_pattern1    = '^"(.+)"\s*(.*)';    # "name with blanks" val
my $file_pattern2    = '^(\S+)\s*(.*)';     # name val

###############################################################################
# all test pattern return :
# an empty array is not match
# an array of 2 values if match
###############################################################################
# low-level sub, to be called by is_macro, is_alias ...
sub test_pattern ($$) {
	my $ligne   = shift(@_);                # string to be tested
	my $pattern = shift(@_);                # motif pattern

	if ( $ligne =~ m/$pattern/ ) {

		#print "found $1 : $2\n";
		return ( $1, $2 );
	}
	else {
		return ();
	}
}
###############################################################################
sub is_macro($) {
	my $ligne = shift(@_);    # string to be tested

	#print "is_macro ";

	return test_pattern( $ligne, $macro_pattern );
}
###############################################################################
sub is_alias($) {
	my $ligne = shift(@_);    # string to be tested

	#print "is_alias ";
	return test_pattern( $ligne, $alias_pattern );
}
###############################################################################
sub is_directive($) {
	my $ligne = shift(@_);    # string to be tested

	#print "is_directive ";
	return test_pattern( $ligne, $directive_pattern );
}
###############################################################################
sub is_file($) {
	my $ligne = shift(@_);    # string to be tested

	if ( $ligne =~ m/$file_pattern1/o ) {

		#print "file pattern1 $1 : $2\n";
		return ( $1, $2 );
	}
	elsif ( $ligne =~ m/$file_pattern2/o ) {

		#print "file pattern2 $1 : $2\n";
		return ( $1, $2 );
	}
	else {
		return ();
	}
}
###############################################################################
sub is_negsel($) {
	my $ligne = shift(@_);    # string to be tested

	#print "is_negsel ";
	return test_pattern( $ligne, $negsel_pattern );
}
###############################################################################
sub is_equalsel($) {
	my $ligne = shift(@_);    # string to be tested

	#print "is_equalsel ";
	my @ret = test_pattern( $ligne, $equalsel_pattern );
	if (@ret) {
		return is_file( shift(@ret) );
	}
	else {
		return ();
	}
}
###############################################################################
sub is_sel($) {
	my $ligne = shift(@_);    # string to be tested

	#print "is_sel ";
	return is_file($ligne);
}
###############################################################################
sub is_anysel($) {
	my $ligne = shift(@_);    # string to be tested

	#print "is_anysel ";
	my @ret = test_pattern( $ligne, $anysel_pattern );
	if (@ret) {
		return is_file( shift(@ret) );
	}
	else {
		return ();
	}
}
###############################################################################
# decode binaries values
# return 1/0 if good, else undef
sub is_binary($) {
	my $val = shift(@_);

	if ( $val =~ m/^(yes|true|on|1)$/ ) {
		return 1;
	}
	elsif ( $val =~ m/^(no|false|off|0)$/ ) {
		return 0;
	}
	else {
		set_error("$val is not a binary value\n");
		return undef;
	}
}
###############################################################################
# check if a number (integer, without sign)
sub is_number($) {
	my $val = shift(@_);

	if ( $val =~ m/^\d+$/ ) {
		return $val;
	}
	else {
		set_error("$val is not a number\n");
		return undef;
	}
}
###############################################################################
# check if it looks like a email adress
sub is_mail_adress($) {
	my $val = shift(@_);

	# a mail adress should contain the "at" character
	if ( $val =~ m/\@/ ) {
		return $val;
	}
	else {
		set_error("$val do not seems to be a mail adress\n");
		return undef;
	}
}
###############################################################################
# check if a number is in [min , max ] interval
sub is_borned_number($$$) {
	my $val = shift(@_);
	my $min = shift(@_);
	my $max = shift(@_);

	if ( !defined is_number($val) ) {
		return undef;
	}
	elsif ( ( $val >= $min ) and ( $val <= $max ) ) {
		return $val;
	}
	else {
		set_error("$val do not seems to be a good value [$min - $max]\n");
		return undef;
	}
}
###############################################################################
# check if it looks like a nice value (man nice on unix)
# we just allow positives value (lower priority)
sub is_nicevalue($) {
	my $val = shift(@_);

	return is_borned_number( $val, 0, 19 );
}
###############################################################################
# check a macro syntaxe
sub check_macro($$$) {
	my $key         = shift(@_);
	my $val         = shift(@_);
	my $allow_undef = shift(@_);

	# the hash contains all available macro key
	# and point to syntaxe checking sub
	my %macro = (
		LINES   => \&is_number,
		MAILTO  => \&is_mail_adress,
		VERBOSE => \&is_binary,
		NICE    => \&is_nicevalue,
		BATCH   => \&is_binary,
	);

	if ( exists $macro{$key} ) {
		if ( ( !defined $val ) or ( $val eq '' ) ) {

			# special case for afickonfig to delete values
			if ($allow_undef) {
				return 1;
			}
			else {
				set_error("missing macro value");
				return undef;
			}
		}
		else {
			return &{ $macro{$key} }($val);
		}
	}
	else {
		set_error("$key is not a valid macro keyword\n");
		return undef;
	}
}
###############################################################################
# check if a file exists
sub is_plainfile($) {
	my $val = shift(@_);

	if ( -f $val ) {
		return $val;
	}
	else {
		set_error("$val plain file does not exists\n");
		return undef;
	}
}
###############################################################################
# check if a file exists
sub check_plainfile($) {
	my $val = shift(@_);

	if ( is_plainfile($val) ) {
		return $val;
	}
	elsif ( open( F, '>', $val ) ) {

		# create file
		close(F);
	}
	else {
		set_error("can not create file $val : $!\n");
		return undef;
	}
}
###############################################################################
# check if a directory exists
sub is_directory($) {
	my $val = shift(@_);

	if ( -d $val ) {
		return $val;
	}
	else {
		set_error("$val directory does not exists\n");
		return undef;
	}
}
###############################################################################
# check if a directory exists
sub check_directory($) {
	my $val = shift(@_);

	if ( is_directory($val) ) {
		return $val;
	}
	elsif ( mkdir( $val, 0700 ) ) {

		# create directory
		return $val;
	}
	else {
		set_error("can not create directory $val : $!\n");
		return undef;
	}
}
###############################################################################
sub is_anyfile($) {
	my $val = shift(@_);

	if ( ( -e $val ) or ( -l $val ) ) {
		return $val;
	}
	else {
		set_error("$val file does not exists\n");
		return undef;
	}
}
###############################################################################
# never return false (undef) : to be used if no checking available
sub is_bidon($) {
	my $val = shift(@_);
	return $val;
}
###############################################################################
# check if it is debug level
sub is_debug_level($) {
	my $val = shift(@_);

	return is_borned_number( $val, 0, 3 );
}
###############################################################################
# check if a valid url
sub is_url($) {
	my $val = shift(@_);

	if ( $val =~ m/^(stdout|stderr|null)$/ ) {
		return $val;
	}
	else {
		set_error("$val is not a valid url\n");
		return undef;
	}
}
###############################################################################
# check if a valid file name (?)
sub is_name($) {
	my $val = shift(@_);

	if ( $val =~ m/^\w+$/ ) {
		return $val;
	}
	else {
		set_error("$val is not a valid name\n");
		return undef;
	}
}
###############################################################################
sub is_format($) {
	my $val = shift(@_);

	my %format = (
		'txt'  => 1,
		'html' => 2,
		'xml'  => 3,
	);

	if ( exists $format{$val} ) {
		return $val;
	}
	else {
		set_error("$val is not a valid format\n");
		return undef;
	}
}
###############################################################################
sub check_directive($$$) {
	my $key         = shift(@_);
	my $val         = shift(@_);
	my $allow_undef = shift(@_);

	my %directive = (
		ignore_case        => \&is_binary,
		report_full_newdel => \&is_binary,
		running_files      => \&is_binary,
		timing             => \&is_binary,
		verbose            => \&is_binary,
		warn_dead_symlinks => \&is_binary,
		warn_missing_file  => \&is_binary,

		max_checksum_size => \&is_number,

		archive => \&check_directory,
		history => \&check_plainfile,

		debug      => \&is_debug_level,
		report_url => \&is_url,
		database   => \&is_bidon,

		exclude_suffix => \&is_bidon,
		exclude_prefix => \&is_bidon,
		exclude_re     => \&is_bidon,

		#output_format  => \&is_format,
	);

	if ( exists $directive{$key} ) {
		if ( ( !defined $val ) or ( $val eq '' ) ) {

			# special case for afickonfig to delete values
			if ($allow_undef) {
				return 1;
			}
			else {
				set_error("missing directive value");
				return undef;
			}
		}
		else {
			return &{ $directive{$key} }($val);
		}
	}
	else {
		set_error("$key is not a directive keyword\n");
		return undef;
	}
}
###############################################################################
# test if a valid operator
sub test_op($) {
	my $op = shift(@_);

	return ( $op =~ m/^[+-]$/ );
}
##########################################################################
# test if a valid attribute
# return decoded value (alias -> attributes) or undef
sub test_attribute($$$) {
	my $val        = shift(@_);
	my $rh_aliases = shift(@_);    # ref to resolved aliases
	my $resolv     = shift(@_);    # binary : resolv ?

	if ( !$resolv ) {
		return $val;
	}
	else {
		return ( exists( $rh_aliases->{$val} ) ? $rh_aliases->{$val} : undef );
	}
}
##########################################################################
# apply an attribute on chain
sub apply_attribute($$$$) {
	my $begin     = shift(@_);     # begin of resolved chain
	my $op        = shift(@_);     # operator
	my $attribute = shift(@_);     # attribute to add/subtract
	my $resolv    = shift(@_);     # binary : resolv ?

	if ( !$resolv ) {
		return "$begin$op$attribute";
	}

	if ( $op eq '+' ) {

		# test duplicate
		if ( $begin =~ m/$attribute/ ) {
			set_error "find duplicate $attribute ($begin)\n";
			return undef;
		}
		else {

			# ok
			$begin .= "$attribute";
		}
	}
	else {

		# negative operator
		if ( $begin =~ s/$attribute// ) {

			# ok
		}
		else {
			set_error "can not remove $attribute from $begin\n";
			return undef;
		}
	}
	return $begin;
}
##########################################################################
{

	# private value used as a cache, to avoid too many calls
	my $sha1_exist;

	sub is_sha1() {
		if ( !defined $sha1_exist ) {
			eval { require Digest::SHA1 };
			if ($@) {
				$sha1_exist = 0;
			}
			else {
				$sha1_exist = 1;
			}
		}
		else {

			# already in cache : nothing to do
		}
		return $sha1_exist;
	}
}
##########################################################################
# check alias syntaxe
# can resolv in base attribute if resolv is 1
# else just check operator syntaxe
# the problem is with afickonfig : when checking arg, we do not know all
# alias defined in config file
#
# warning : this version stop at first error
#   and do not try to skip bad elements
sub check_alias($$$) {
	my $var         = shift(@_);    # string containing the alias to be resolved
	my $rh_aliases  = shift(@_);    # ref to hash containing aliases
	my $allow_undef = shift(@_);

	if ( ( !defined $var ) or ( $var eq '' ) ) {
		if ($allow_undef) {
			return 1;
		}
		else {
			set_error("missing alias value");
			return undef;
		}
	}

	# resolv is set from alias hash
	my $resolv = defined $rh_aliases;    # a binary value

	remove_all_spaces( \$var );

	my @tab = split( /\b/, $var );

	# first attribute
	my $first = shift(@tab);

	# test resolution
	my $real = test_attribute( $first, $rh_aliases, $resolv );
	if ( !defined $real ) {
		set_error "$first is not a valid attribute\n";
		return undef;
	}

	while ( my $op = shift(@tab) ) {

		if ( test_op($op) ) {
			my $next = shift(@tab);
			if ( !defined $next ) {
				set_error "missing last attribute\n";
				return undef;
			}
			my $next2 = test_attribute( $next, $rh_aliases, $resolv );
			if ( defined $next2 ) {
				$real = apply_attribute( $real, $op, $next2, $resolv );
				return undef if ( !defined $real );
			}
			else {
				set_error "$next is not a valid attribute\n";
				return undef;
			}
		}
		else {
			set_error "$op should be the operator + or -\n";
			return undef;
		}
	}

	# test for checksum :
	# in current version, only one checksum is available, md5 by default
	# and use of sha1 override md5
	if ( $real =~ m/1/ ) {
		if ( is_sha1() ) {
			if ( $real =~ m/5/ ) {

				# remove md5
				$real =~ s/5//;
			}
		}
		elsif ( $real =~ m/5/ ) {

			# sha1 not available, remove it
			$real =~ s/1//;
		}
		else {

			# sha1 not available, replaced by md5
			$real =~ s/1/5/;
		}
	}

	debug("check_alias : $var -> $real\n");
	return $real;
}
##########################################################################
# check duplicate for alias, macros
sub check_duplicate($$) {
	my $key     = shift(@_);
	my $rh_find = shift(@_);

	debug("check_duplicate $key\n");

	if ( $key =~ m/^exclude/ ) {

		# set to 1 on first call
		$rh_find->{$key} = 1 if ( !exists( $rh_find->{$key} ) );
		return 0;
	}
	elsif ( exists( $rh_find->{$key} ) ) {
		set_error("find duplicate key\n");
		return 1;
	}
	else {
		$rh_find->{$key}++;
		return 0;
	}
}
##########################################################################
sub get_configuration($$$$$$) {
	my $configfile   = shift(@_);
	my $rh_macro     = shift(@_);
	my $rh_alias     = shift(@_);
	my $rh_directive = shift(@_);
	my $rh_rule      = shift(@_);
	my $rh_onlydir   = shift(@_);

	my @config;
	read_config( $configfile, \@config ) or die get_error();

	# some hash to detect multiple lines
	my %h_m_macro;
	my %h_m_directive;
	my %h_m_alias;
	my %h_m_rule;

	my $nb_pbs  = 0;
	my $line_id = 0;
	foreach my $line (@config) {
		chomp $line;

		$line =~ s/#.*//;     # skip comments
		                      #remove_trailing_spaces( \$line );
		$line =~ s?\\?/?g;    # replace \ by / (for windows)

		next unless length($line);    # skip blank lines

		#debug( "config line = $line\n", 1);

		my @ret;
		if ( @ret = is_macro($line) ) {

			# macros lines
			###################
			my $key = shift(@ret);
			my $val = shift(@ret);

			if (   ( !defined check_macro( $key, $val, 0 ) )
				or ( check_duplicate( $key, \%h_m_macro ) ) )
			{
				$nb_pbs++;
				next;
			}
			debug( "found macro $key : $val\n", 2 );

			# add to Macro for print_config
			$rh_macro->{$key} = $val;
		}
		elsif ( @ret = is_directive($line) ) {

			# a configuration line
			######################
			my $key = shift(@ret);
			my $val = shift(@ret);
			remove_trailing_spaces( \$val );

			my $ret2 = check_directive( $key, $val, 0 );
			if (   ( !defined $ret2 )
				or ( check_duplicate( $key, \%h_m_directive ) ) )
			{
				$nb_pbs++;
				next;
			}
			$rh_directive->{$key} = $val;
		}
		elsif ( @ret = is_alias($line) ) {

			# is an alias definition
			########################
			my $key = shift(@ret);
			my $val = shift(@ret);
			remove_all_spaces( \$val );

			my $decoded = check_alias( $val, $rh_alias, 0 );
			if (   ( !defined $decoded )
				or ( check_duplicate( $key, \%h_m_alias ) ) )
			{
				$nb_pbs++;
				next;
			}

			$rh_alias->{$key} = $decoded;

			debug( "alias $key -> $decoded\n", 2 );
		}
		elsif ( @ret = is_negsel($line) ) {

			my $key = shift(@ret);
			$rh_directive->{$key} = 0;

		}
		elsif ( @ret = is_equalsel($line) ) {
			my $key       = shift(@ret);
			my $attribute = shift(@ret);

			my $masq = check_alias( $attribute, $rh_alias, 0 );
			if ( !defined $masq ) {
				$nb_pbs++;
				next;
			}

			$rh_rule->{$key}    = $attribute;
			$rh_onlydir->{$key} = 1;

		}
		elsif ( @ret = is_sel($line) ) {
			my $key       = shift(@ret);
			my $attribute = shift(@ret);

			my $masq = check_alias( $attribute, $rh_alias, 0 );
			if ( !defined $masq ) {
				$nb_pbs++;
				next;
			}
			$rh_rule->{$key} = $attribute;

		}
		else {
			warning(
				"skipped config file $line (line $line_id) : unknown type\n");
		}
	}    # foreach my $line

	return $nb_pbs;
}
##########################################################################
return 1;
